from .lookup import Lookup
from .rule import LookupRules
from .exceptions import MissingMandatoryValue
from peerplays.event import Event
from peerplays.rule import Rule
from peerplays.asset import Asset
from peerplays.bettingmarketgroup import (
    BettingMarketGroups, BettingMarketGroup)


class LookupBettingMarketGroup(Lookup, dict):
    """ Lookup Class for betting market groups

        :param dict bmg: Lookup content (files) for the BMG
        :param LookupEvent event: Parent LookupEvent for BMG
        :param dict extra_data: Optionally provide additional data that is
               stored in the same dictionary

    """

    operation_update = "betting_market_group_update"
    operation_create = "betting_market_group_create"

    def __init__(
        self,
        bmg,
        event,
        extra_data={}
    ):
        Lookup.__init__(self)
        self.identifier = "{}/{}".format(
            event.names_json["en"],
            bmg["description"]["en"]
        )
        self.event = event
        self.parent = event
        dict.__init__(self, extra_data)
        dict.update(
            self,
            bmg
        )
        for mandatory in [
            "description",
            "asset",
            "bettingmarkets",
            "rules",
        ]:
            if mandatory not in self:
                raise MissingMandatoryValue(
                    "A value for '{}' is mandatory".format(
                        mandatory
                    )
                )

        # FIXME: Figure out if the name has a variable
        # FIXME: Figure out if this is a dynamic bmg

    @property
    def sport(self):
        """ Return the sport for this BMG
        """
        return self.parent.sport

    @property
    def rules(self):
        """ Return instance of LookupRules for this BMG
        """
        assert self["rules"] in self.sport["rules"]
        return LookupRules(self.sport["identifier"], self["rules"])

    def test_operation_equal(self, bmg):
        """ This method checks if an object or operation on the blockchain
            has the same content as an object in the  lookup
        """
        def is_update(bmg):
            return any([x in bmg for x in [
                "betting_market_group_id", "new_description",
                "new_event_id", "new_rules_id"]])

        def is_create(bmg):
            return any([x in bmg for x in [
                "description", "event_id", "rules_id"]])

        if not is_create(bmg) and not is_update(bmg):
            raise ValueError

        lookupdescr = self.description
        chainsdescr = [[]]
        prefix = "new_" if is_update(bmg) else ""
        chainsdescr = bmg[prefix + "description"]
        rules_id = bmg[prefix + "rules_id"]
        event_id = bmg[prefix + "event_id"]
        if is_update(bmg):
            frozen = bmg["frozen"]
            delay_bets = bmg["delay_bets"]
        else:
            frozen = False
            delay_bets = False

        # Test if Rules and Events exist
        # only if the id starts with 1.
        test_rule = rules_id[0] == 1
        if test_rule:
            Rule(rules_id)

        test_event = event_id[0] == 1
        if test_event:
            Event(event_id)

        if (
            all([a in chainsdescr for a in lookupdescr]) and
            all([b in lookupdescr for b in chainsdescr]) and
            (not test_event or event_id == self.event.id) and
            (not test_rule or rules_id == self.rules.id) and
            (self.get("frozen", False) == frozen) and
            (self.get("delay_bets", False) == delay_bets)
        ):
            return True
        return False

    def find_id(self):
        """ Try to find an id for the object of the  lookup on the
            blockchain

            ... note:: This only checks if a sport exists with the same name in
                       **ENGLISH**!
        """
        # In case the parent is a proposal, we won't
        # be able to find an id for a child
        parent_id = self.parent.id
        if parent_id[0] == "0" or parent_id[:4] == "1.10":
            return

        bmgs = BettingMarketGroups(
            self.parent.id,
            peerplays_instance=self.peerplays)
        en_descrp = next(filter(lambda x: x[0] == "en", self.description))

        for bmg in bmgs:
            if en_descrp in bmg["description"]:
                return bmg["id"]

    def is_synced(self):
        """ Test if data on chain matches lookup
        """
        if "id" in self:
            bmg = BettingMarketGroup(self["id"])
            if self.test_operation_equal(bmg):
                return True
        return False

    def propose_new(self):
        """ Propose operation to create this object
        """
        asset = Asset(
            self["asset"],
            peerplays_instance=self.peerplays)
        return self.peerplays.betting_market_group_create(
            self.description,
            event_id=self.event.id,
            rules_id=self.rules.id,
            asset=asset["id"],
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    def propose_update(self):
        """ Propose to update this object to match  lookup
        """
        return self.peerplays.betting_market_group_update(
            self.id,
            self.description,
            event_id=self.event.id,
            rules_id=self.rules.id,
            frozen=self.get("frozen", False),
            delay_bets=self.get("delay_bets", False),
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    @property
    def bettingmarkets(self):
        """ Return instances of LookupBettingMarket for this BMG
        """

        from .bettingmarket import LookupBettingMarket

        # Allow to overwrite the variables that might be in the betting market
        # definition (such as home team and away team names)
        class Teams:
            home = " ".join([
                x.capitalize() for x in self.event["teams"][0].split(" ")])
            away = " ".join([
                x.capitalize() for x in self.event["teams"][1].split(" ")])

        for market in self["bettingmarkets"]:
            description = dict()

            # Overwrite the description with with proper replacement of variables
            for k, v in market["description"].items():
                description[k] = v.format(
                    teams=Teams
                )
            yield LookupBettingMarket(
                description=description,
                bmg=self
            )

    @property
    def description(self):
        """ Properly format description for internal use
        """
        return [
            [
                k,
                v
            ] for k, v in self["description"].items()
        ]



import logging
from colorlog import ColoredFormatter

LOG_LEVEL = logging.DEBUG
LOGFORMAT = ("  %(log_color)s%(levelname)-8s%(reset)s |"
             " %(log_color)s%(message)s%(reset)s")
logging.root.setLevel(LOG_LEVEL)
formatter = ColoredFormatter(LOGFORMAT)
stream = logging.StreamHandler()
stream.setLevel(LOG_LEVEL)
stream.setFormatter(formatter)
log = logging.getLogger(__name__)
log.setLevel(LOG_LEVEL)
log.addHandler(stream)

# logging.basicConfig(level=logging.DEBUG)

UPDATE_PROPOSING_NEW = 1
UPDATE_PENDING_NEW = 2



from .lookup import Lookup


class LookupParticipants(Lookup, dict):
    """ Lookup Class for participants

        :param str sport: Identifier for sport
        :param str list_identifier: Identifier for the list of participants
    """

    operation_update = ""
    operation_create = ""

    def __init__(self, sport, list_identifier):
        self.identifier = "{}/{}".format(sport, list_identifier)
        super(LookupParticipants, self).__init__()
        assert sport in self.data["sports"], "Sport {} not avaialble".format(
            sport
        )

        if list_identifier in self.data["sports"][sport]["participants"]:
            dict.__init__(
                self,
                self.data["sports"][sport]["participants"][list_identifier]
            )

    def is_participant(self, name):
        """ See if we can find a particular participant in the list of
            participants (self)
        """
        parties = self.get("participants")
        for team in parties:
            if (
                name.lower() in [
                    x.lower() for x in team.get("name", {}).values()] or
                name.lower() in [
                    x.lower() for x in team.get("aliases", [])]
            ):
                return True
        return False



from .lookup import Lookup
from .rule import LookupRules
from peerplays.bettingmarket import (
    BettingMarket, BettingMarkets)
from peerplays.bettingmarketgroup import (
    BettingMarketGroup, BettingMarketGroups)


class LookupBettingMarket(Lookup, dict):
    """ Lookup Class for Betting Market

        :param dict description: list of internationalized descriptions
        :param LookupBettingMarketGroup bmg: Parent element
        :param dict extra_data: Optionally provide additional data that is
               stored in the same dictionary

    """
    operation_update = "betting_market_update"
    operation_create = "betting_market_create"

    def __init__(
        self,
        description,
        bmg,
        extra_data={}
    ):
        Lookup.__init__(self)
        self.identifier = "{}/{}".format(
            bmg["description"]["en"],
            description["en"]
        )
        self.bmg = bmg
        self.parent = bmg
        dict.__init__(self, extra_data)
        dict.update(
            self, {
                "description": description
            }
        )
        # FIXME: Figure out if the name has a variable
        # FIXME: Figure out if this is a dynamic bmg

    @property
    def event(self):
        """ Return parent Event
        """
        return self.parent.event

    @property
    def group(self):
        """ Return parent BMG
        """
        return self.parent

    def test_operation_equal(self, bmg):
        """ This method checks if an object or operation on the blockchain
            has the same content as an object in the  lookup
        """
        def is_update(bmg):
            return any([x in bmg for x in [
                "new_group_id", "new_description",
                "betting_market_id"]])

        def is_create(bmg):
            return any([x in bmg for x in [
                "group_id", "description"]])

        if not is_create(bmg) and not is_update(bmg):
            raise ValueError

        lookupdescr = self.description
        chainsdescr = [[]]
        prefix = "new_" if is_update(bmg) else ""
        chainsdescr = bmg[prefix + "description"]
        group_id = bmg[prefix + "group_id"]

        test_group = group_id[0] == 1
        if test_group:
            BettingMarketGroup(group_id)

        if (all([a in chainsdescr for a in lookupdescr]) and
                all([b in lookupdescr for b in chainsdescr]) and
                (not test_group or group_id == self.group.id)):
            return True
        return False

    def find_id(self):
        """ Try to find an id for the object of the  lookup on the
            blockchain

            ... note:: This only checks if a sport exists with the
                        same description in **ENGLISH**!
        """
        # In case the parent is a proposal, we won't
        # be able to find an id for a child
        parent_id = self.parent.id
        if parent_id[0] == "0" or parent_id[:4] == "1.10":
            return

        bms = BettingMarkets(
            self.parent.id,
            peerplays_instance=self.peerplays)
        en_descrp = next(filter(lambda x: x[0] == "en", self.description))

        for bm in bms:
            if en_descrp in bm["description"]:
                return bm["id"]

    def is_synced(self):
        """ Test if data on chain matches lookup
        """
        if "id" in self:
            bmg = BettingMarket(self["id"])
            if self.test_operation_equal(bmg):
                return True
        return False

    def propose_new(self):
        """ Propose operation to create this object
        """
        return self.peerplays.betting_market_create(
            description=self.description,
            payout_condition=[],
            group_id=self.parent.id,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    def propose_update(self):
        """ Propose to update this object to match  lookup
        """
        return self.peerplays.betting_market_update(
            self.id,
            payout_condition=[],
            description=self.description,
            group_id=self.parent.id,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    @property
    def description(self):
        """ This method ensures that the description has the proper format as
            well as the proper string replacements for teams
        """
        return [
            [
                k,
                v.format(**self)   # replace variables
            ] for k, v in self["description"].items()
        ]



from .lookup import Lookup
from peerplays.rule import Rules, Rule


class LookupRules(Lookup, dict):
    """ Lookup Class for Rule

        :param str sport: Sport Identifier
        :param str rules: Rules Identifier
    """

    operation_update = "betting_market_rules_update"
    operation_create = "betting_market_rules_create"

    def __init__(self, sport, rules):
        self.identifier = "{}/{}".format(sport, rules)
        super(LookupRules, self).__init__()
        assert sport in self.data["sports"], "Sport {} not avaialble".format(
            sport
        )
        assert rules in self.data["sports"][sport]["rules"], \
            "rules {} not avaialble in sport {}".format(
                rules, sport)
        # This is a list and not a dictionary!
        dict.__init__(
            self,
            self.data["sports"][sport]["rules"][rules]
        )

    def test_operation_equal(self, operation):
        """ This method checks if an object or operation on the blockchain
            has the same content as an object in the  lookup
        """
        lookupnames = self.descriptions
        chainsnames = [[]]
        if "name" in operation:
            chainsnames = operation["description"]
        elif "new_name" in operation:
            chainsnames = operation["new_description"]
        else:
            raise ValueError

        if (all([a in chainsnames for a in lookupnames]) and
                all([b in lookupnames for b in chainsnames])):
            return True

    def find_id(self):
        """ Try to find an id for the object of the  lookup on the
            blockchain

            ... note:: This only checks if a sport exists with the same name in
                       **ENGLISH**!
        """
        rules = Rules(peerplays_instance=self.peerplays)
        for rule in rules:
            if (
                ["en", self["name"]["en"]] in rule["name"]
            ):
                return rule["id"]

    def is_synced(self):
        """ Test if data on chain matches lookup
        """
        if "id" in self:
            sport = Rule(self["id"])
            if self.test_operation_equal(sport):
                return True
        return False

    def propose_new(self):
        """ Propose operation to create this object
        """
        return self.peerplays.betting_market_rules_create(
            self.names,
            self.descriptions,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    def propose_update(self):
        """ Propose to update this object to match  lookup
        """
        return self.peerplays.betting_market_rules_update(
            self["id"],
            names=self.names,
            descriptions=self.descriptions,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    @property
    def names(self):
        """ Properly format names for internal use
        """
        return [
            [
                k,
                v
            ] for k, v in self["name"].items()
        ]

    @property
    def descriptions(self):
        """ Properly format descriptions for internal use
        """
        return [
            [
                k,
                v
            ] for k, v in self["description"].items()
        ]



from .lookup import Lookup
from .rule import LookupRules
from .exceptions import MissingMandatoryValue
from peerplays.event import Event
from peerplays.rule import Rule
from peerplays.asset import Asset
from peerplays.bettingmarketgroup import (
    BettingMarketGroups, BettingMarketGroup)


class LookupBettingMarketGroupResolve(Lookup, dict):
    """ Lookup Class for Resolving BettingMarketGroups
    """

    operation_update = None
    operation_create = "betting_market_group_resolve"

    def __init__(
        self,
        bmg,
        result,
        extra_data={}
    ):
        Lookup.__init__(self)
        self.identifier = "{}/result".format(
            bmg["description"]["en"],
        )
        self.parent = bmg
        dict.__init__(self, extra_data)
        dict.update(
            self,
            bmg
        )
        assert isinstance(result, list) and len(result) == 2, \
            "Result must be a list of length 2."
        dict.update(self, {
            "result": result
        })

    @property
    def bmg(self):
        """ The BMG is the parent
        """
        return self.parent

    @property
    def markets(self):
        """ The BMG is the parent
        """
        return self.parent.bettingmarkets

    @property
    def sport(self):
        """ Return the sport for this BMG
        """
        return self.parent.sport

    @property
    def rules(self):
        """ Return instance of LookupRules for this BMG
        """
        assert self["rules"] in self.sport["rules"]
        return LookupRules(self.sport["identifier"], self["rules"])

    @property
    def resolutions(self):
        """ This property constructs the resultions array to be used in the
            transactions. It takes the following form

            ... code-block: js

                [
                    ["1.21.257", "win"],
                    ["1.21.258", "not_win"],
                    ["1.21.259", "cancel"],
                ]


        """
        grading = self.rules.get("grading")
        assert grading, "Rules {} have no grading!?".format(self.rules["identifier"])
        assert "metric" in grading
        assert "resolutions" in grading

        # Define variables we want to use when grading

        class Result:
            hometeam = (self["result"][0])
            awayteam = float(self["result"][1])
            total = sum([float(x) for x in self["result"]])

            # aliases
            home = hometeam
            away = awayteam

        def return_metric(s):
            if not isinstance(s, str):
                raise ValueError(
                    "metric must be string, was {}".format(
                        type(s)
                    ))
            try:
                metric = eval(s.format(result=Result))
            except Exception:
                raise Exception("Cannot evaluate metric '{}' -> '{}'".format(
                    s, s.format(result=Result)))
            return metric

        def evaluate_metric(equation, metric):
            if not isinstance(equation, str):
                raise ValueError(
                    "equation must be string, was {}".format(
                        type(equation)
                    ))
            try:
                metric = eval(equation.format(metric=metric))
            except Exception:
                raise Exception("Cannot evaluate metric '{}' -> '{}'".format(
                    equation, equation.format(result=Result)))
            return metric

        metric = return_metric(grading.get("metric", ""))
        bettingmarkets = self.markets
        ret = []
        for market in grading.get("resolutions", []):
            bettingmarket = next(bettingmarkets)
            resolved = {
                key: evaluate_metric(option, metric)
                for key, option in market.items()
            }
            # The resolved dictionary looks like this
            # {'win': False, 'not_win': True, 'void': False}
            # we now need to ensure that only one of those options is 'true'
            assert sum(resolved.values()) == 1, \
                "Multiple options resolved to 'True': {}".format(
                    str(resolved))

            ret.extend([
                [bettingmarket.id, key]
                for key, value in resolved.items() if value
            ])

        return ret

    def test_operation_equal(self, resolve):
        """ This method checks if an object or operation on the blockchain
            has the same content as an object in the  lookup
        """

        lookupresults = self.resolutions
        chainsresults = resolve["resolutions"]
        bmg_id = resolve["betting_market_group_id"]

        # Test if BMG exists
        # only if the id starts with 1.
        test_bmg = bmg_id[0] == 1
        if test_bmg:
            BettingMarketGroup(bmg_id)

        if (
            all([a in chainsresults for a in lookupresults]) and
            all([b in lookupresults for b in chainsresults]) and
            (not test_bmg or bmg_id == self.parent.id)
        ):
            return True
        return False

    def find_id(self):
        """ Market resolve operations do not have their own ids
        """
        pass

    def is_synced(self):
        """ Here, we need to figure out if the market has already been resolved
        """
        # FIXME  / TODO
        return False

    def propose_new(self):
        """ This call proposes the resolution of the betting market group
        """
        return self.peerplays.betting_market_resolve(
            self.parent.id,
            self.resolutions,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    def propose_update(self):
        """ There is no such thing as an updated resolution
        """
        pass



import sys
import os
import yaml
from peerplays.instance import shared_peerplays_instance
from peerplays.account import Account
from peerplays.proposal import Proposal, Proposals
from peerplays.storage import configStorage as config
from . import log, UPDATE_PENDING_NEW, UPDATE_PROPOSING_NEW
from .exceptions import ObjectNotFoundError
from bookiesports import BookieSports


class LookupDatabaseConfig:
    pass


class Lookup(dict):

    #: Singelton to store data and prevent rereading if Lookup is
    #: instantiated multiple times
    data = dict()
    approval_map = {}

    direct_buffer = None
    proposal_buffer = None
    sports_folder = None

    def __init__(
        self,
        sports_folder=None,
        peerplays_instance=None,
        proposing_account=None,
        approving_account=None,
        *args,
        **kwargs
    ):
        """ Let's load all the data from the folder and its subfolders
        """
        self.peerplays = peerplays_instance or shared_peerplays_instance()
        # self._cwd = os.path.dirname(os.path.realpath(__file__))
        self._cwd = os.getcwd()

        if not proposing_account:
            if "default_account" in config:
                proposing_account = config["default_account"]
            else:
                proposing_account = proposing_account
        self.proposing_account = proposing_account

        if not approving_account:
            if "default_account" in config:
                approving_account = config["default_account"]
            else:
                approving_account = approving_account
        self.approving_account = approving_account

        # We define two transaction buffers
        if not Lookup.direct_buffer:
            self.clear_direct_buffer()
        if not Lookup.proposal_buffer:
            self.clear_proposal_buffer()

        # Do not reload sports if already stored in data
        if (
            not Lookup.data or
            Lookup.sports_folder != sports_folder
        ):
            # Load sports
            self.data["sports"] = BookieSports(sports_folder)
            Lookup.sports_folder = sports_folder

    def set_approving_account(self, account):
        self.approving_account = account

    def set_proposing_account(self, account):
        self.proposing_account = account

    def pack(self):
        return dict(
            data=dict(self.data),
            approval_map=self.approval_map,
            approving_account=self.approving_account,
            proposing_account=self.proposing_account
        )

    @classmethod
    def unpack(cls, pack):
        inst = cls()
        dict.__init__(inst, pack.get("data", {}))
        inst.approval_map = pack["approval_map"]
        inst.approving_account = pack["approving_account"]
        inst.proposing_account = pack["proposing_account"]
        return inst

    @property
    def wallet(self):
        return self.peerplays.wallet

    @staticmethod
    def _clear():
        Lookup.data = dict()
        Lookup.approval_map = {}
        Lookup.direct_buffer = None
        Lookup.proposal_buffer = None

    def clear_proposal_buffer(self):
        Lookup.proposal_buffer_tx = self.peerplays.new_tx()
        Lookup.proposal_buffer = self.peerplays.new_proposal(
            Lookup.proposal_buffer_tx,
            proposer=self.proposing_account)

    def clear_direct_buffer(self):
        Lookup.direct_buffer = self.peerplays.new_tx()

    def broadcast(self):
        """ Since we are using multiple txbuffers, we need to do multiple
            broadcasts
        """
        from pprint import pprint
        pprint(Lookup.direct_buffer.broadcast())
        pprint(Lookup.proposal_buffer.broadcast())

    def _loadyaml(self, f):
        """ Load a YAML file

            :param str f: YAML File location
        """
        try:
            with open(f) as fid:
                t = yaml.load(fid)
            return t
        except yaml.YAMLError as exc:
            log.error("Error in configuration file {}: {}".format(f, exc))
            sys.exit(1)
        except Exception as e:
            log.error("The file {} is required but doesn't exist! ({})".format(
                f, e))
            sys.exit(1)

    # List calls
    def list_sports(self):
        """ List all sports in the  lookup
        """
        from .sport import LookupSport
        return [LookupSport(x) for x in self.data["sports"]]

    def get_sport(self, sportname):
        from .sport import LookupSport
        return LookupSport(sportname)

    # Update call
    def update(self):
        """ This call makes sure that the data in the  lookup matches
            the data on the blockchain for the object we are currenty looking
            at.

            It works like this:

            1. Test if the  lookup knows the "id" of the object on chain
            2. If it does not, try to identify the object from the blockchain
                * if available, warn about existing id
                * if pending creation proposal, approve it
                * if none of the above, create proposal
            3. Test if  lookup and blockchain data match, if not
                * if exists proposal for update, approve
                * if not, create proposal to update
        """
        # See if  lookup already has an id
        if "id" not in self or not self["id"]:

            # Test if an object with the characteristics (i.e. name) exist
            id = self.find_id()
            has_pending_new = self.has_pending_new()
            if id:
                log.error((
                    "Object \"{}\" carries id {} on the blockchain. "
                    "Please update your lookup"
                ).format(self.identifier, id))
                self["id"] = id
            elif has_pending_new:
                log.warn((
                    "Object \"{}\" has pending update proposal. Approving ..."
                ).format(self.identifier))
                self.approve(*has_pending_new)
                return UPDATE_PENDING_NEW
            else:
                log.warn((
                    "Object \"{}\" does not exist on chain. Proposing ..."
                ).format(self.identifier))
                self.propose_new()
                return UPDATE_PROPOSING_NEW

        if not self.is_synced():
            log.warn("Object not fully synced: {}: {}".format(
                self.__class__.__name__,
                str(self.get("name", ""))
            ))
            has_pending_update = self.has_pending_update()
            if has_pending_update:
                log.info("Object has pending update: {}: {} in {}".format(
                    self.__class__.__name__,
                    str(self.get("name", "")),
                    str(has_pending_update)
                ))
                self.approve(*has_pending_update)
            else:
                log.info("Object has no pending update, yet: {}: {}".format(
                    self.__class__.__name__,
                    str(self.get("name", ""))
                ))
                self.propose_update()

    def get_pending_operations(self, account="witness-account"):
        pending_proposals = Proposals(account)
        ret = []
        for proposal in pending_proposals:
            if not proposal["id"] in Lookup.approval_map:
                Lookup.approval_map[proposal["id"]] = {}
            for oid, operations in enumerate(proposal.proposed_operations):
                if oid not in Lookup.approval_map[proposal["id"]]:
                    Lookup.approval_map[proposal["id"]][oid] = False
                ret.append((operations, proposal["id"], oid))
        return ret

    def get_buffered_operations(self):
        # Obtain the proposals that we have in our buffer
        # from peerplaysbase.operationids import getOperationNameForId
        for oid, op in enumerate(
            Lookup.proposal_buffer.list_operations()
        ):
            yield op.json(), "0.0.0", "0.0.%d" % oid

    def approve(self, pid, oid):
        """ Approve a proposal

            This call basically flags a single update operation of a proposal
            as "approved". Only if all operations in the proposal are approved,
            will this tool approve the whole proposal and otherwise ignore the
            proposal.

            The call has to identify the correct operation of a proposal on its
            own.

            Internally, a proposal is approved partially using a map that
            contains the approval of each operation in a proposal. Once all
            operations of a proposal are approved, the whole proopsal is
            approved.

            :param str pid: Proposal id
            :param int oid: Operation number within the proposal
        """
        if pid[:3] == "0.0":
            log.info("Cannot approve pending-for-broadcast proposals")
            return
        Lookup.approval_map[pid][oid] = True
        approved_read_for_delete = []
        for p in Lookup.approval_map:
            if all(Lookup.approval_map[p].values()):
                proposal = Proposal(p)
                account = Account(self.approving_account)
                if account["id"] not in proposal["available_active_approvals"]:
                    log.info("Approving proposal {}".format(p))
                    approved_read_for_delete.append(p)
                    self.peerplays.approveproposal(
                        p,
                        account=self.approving_account,
                        append_to=Lookup.direct_buffer
                    )
                else:
                    log.info(
                        "Proposal {} has already been approved by us".format(p)
                    )

        log.info("Approval Map: {}".format(str(Lookup.approval_map)))

        # In order not to approve the same proposal again and again, we remove
        # it from the map
        for p in approved_read_for_delete:
            del Lookup.approval_map[p]

    def has_pending_new(self):
        """ This call tests if a proposal that would create this object is
            pending on-chain

            It only returns true if the exact content is proposed
        """
        from peerplaysbase.operationids import getOperationNameForId
        for op, pid, oid in self.get_pending_operations():
            if getOperationNameForId(op[0]) == self.operation_create:
                if self.test_operation_equal(op[1]):
                    return pid, oid

    def has_buffered_new(self):
        """ This call tests if an operation is buffered for proposal

            It only returns true if the exact content is proposed
        """
        from peerplaysbase.operationids import getOperationNameForId
        for op, pid, oid in self.get_buffered_operations():
            if getOperationNameForId(op[0]) == self.operation_create:
                if self.test_operation_equal(op[1]):
                    return pid, oid

    def has_pending_update(self):
        """ Test if there is an update on-chain to properly match blockchain
            content with lookup content

            It only returns true if the exact content is proposed
        """
        from peerplaysbase.operationids import getOperationNameForId
        for op, pid, oid in self.get_pending_operations():
            if getOperationNameForId(op[0]) == self.operation_update:
                if self.test_operation_equal(op[1]):
                    return pid, oid

    def has_buffered_update(self):
        """ Test if there is an update buffered locally to properly match
            blockchain content with  lookup content

            It only returns true if the exact content is proposed
        """
        from peerplaysbase.operationids import getOperationNameForId
        for op, pid, oid in self.get_buffered_operations():
            if getOperationNameForId(op[0]) == self.operation_update:
                if self.test_operation_equal(op[1]):
                    return pid, oid

    @property
    def id(self):
        """ Returns the id of the object on chain

            :raises IdNotFoundError: if the object couldn't be matched to an
                object on chain
        """
        # Do we already know the id?
        if (
            "id" in self and
            self["id"] and
            isinstance(self["id"], str) and
            len(self["id"].split(".")) == 3
        ):
            return self["id"]

        # Try find the id on the blockchain
        found = self.find_id()
        if found:
            return found

        # Try find the id in the pending on-chain proposals
        found = self.has_pending_new()
        if found:
            return found[0]

        # Try find the id in the locally buffered proposals
        found = self.has_buffered_new()
        if found:
            return found[1]

        raise ObjectNotFoundError(
            "Object not found on chain: {}: {}".format(
                self.__class__.__name__,
                str(self.items())))
        return found

    @property
    def parent_id(self):
        if hasattr(self, "parent"):
            return self.parent.id

    # Prototypes #############################################################
    def test_operation_equal(self, sport):
        """ This method checks if an object or operation on the blockchain
            has the same content as an object in the  lookup
        """
        pass

    def find_id(self):
        """ Try to find an id for the object of the  lookup on the
            blockchain

            ... note:: This only checks if a sport exists with the same name in
                       **ENGLISH**!
        """
        pass

    def is_synced(self):
        """ Test if data on chain matches lookup
        """
        # Compare blockchain content with  lookup

    def propose_new(self):
        """ Propose operation to create this object
        """
        pass

    def propose_update(self):
        """ Propose to update this object to match  lookup
        """
        pass



import datetime
from .lookup import Lookup
from .sport import LookupSport
from .eventgroup import LookupEventGroup
from .bettingmarketgroup import LookupBettingMarketGroup
from peerplays.event import Event, Events
from peerplays.utils import formatTime, parse_time
# from . import log


def subsitution(teams, scheme):
    class Teams:
        home = " ".join([
            x.capitalize() for x in teams[0].split(" ")])
        away = " ".join([
            x.capitalize() for x in teams[1].split(" ")])

    ret = dict()
    for lang, name in scheme.items():
        ret[lang] = name.format(
            teams=Teams
        )
    return ret


class LookupEvent(Lookup, dict):
    """ Lookup Class for an Event

        :param list teams: Teams (first element is **HomeTeam**)
        :param str eventgroup_identifier: Identifier of the event group
        :param str sport_identifier: Identifier of the sport
        :param list season: Internationalized string for the season
        :param datetime.datetime start_time: Datetime the event starts
               (required when creating an event)
        :param dict extra_data: Optionally provide additional data that is
               stored in the same dictionary

        ... note:: Please note that the list of teams begins with the **home**
                   team! Only two teams per event are supported!
    """

    operation_update = "event_update"
    operation_create = "event_create"

    def __init__(
        self,
        teams,
        eventgroup_identifier,
        sport_identifier,
        season={},
        start_time=None,
        id=None,
        extra_data={},
        **kwargs
    ):
        Lookup.__init__(self)

        # Also store all the stuff in kwargs
        dict.__init__(self, extra_data)
        dict.update(self, {
            "teams": teams,
            "eventgroup_identifier": eventgroup_identifier,
            "sport_identifier": sport_identifier,
            "season": season,
            "start_time": start_time,
            "id": id})

        # Define "id" if not present
        self["id"] = self.get("id", None)

        if not len(self["teams"]) == 2:
            raise ValueError(
                "Only matches with two players are allowed! "
                "Here: {}".format(str(self["teams"]))
            )

        self.parent = self.eventgroup
        self.identifier = "{}/{}/{}".format(
            self.parent["name"]["en"],
            teams[0],
            teams[1])

        if start_time and not isinstance(
            self["start_time"], datetime.datetime
        ):
            raise ValueError(
                "'start_time' must be instance of datetime.datetime()")

        if not isinstance(self["season"], dict):
            raise ValueError(
                "'season' must be (language) dictionary")

        if not self.test_teams_valid():
            raise ValueError(
                "Team names not known: {}".format(
                    str(self["teams"])))

        # Initialize name key
        dict.update(self, dict(name=self.names_json))

    def test_teams_valid(self):
        return all(
            self.participants.is_participant(t)
            for t in self["teams"]
        )

    @classmethod
    def find_event(
        cls,
        sport_identifier,
        eventgroup_identifier,
        teams,
        start_time
    ):
        """ This class method is used to find an event by providing:

            :param str sport_identifier: Identifier string for the sport
            :param str eventgroup_identifier: Identifier string for the eventgroup/league
            :param list teams: list of teams
            :param datetime.datetime start_time: Time of start

        """
        sport = LookupSport(sport_identifier)
        eventgroup = LookupEventGroup(sport, eventgroup_identifier)
        events = Events(eventgroup.id)  # This is a pypeerplays class!
        # Format teams into proper names according to event scheme
        names = subsitution(teams, eventgroup["eventscheme"]["name"])
        names = [[k, v] for k, v in names.items()]
        for event in events:
            if (
                any([x in event["name"] for x in names]) and
                formatTime(start_time) == event["start_time"]
            ):
                return cls(
                    id=event["id"],
                    teams=teams,
                    eventgroup_identifier=eventgroup_identifier,
                    sport_identifier=sport_identifier,
                    start_time=start_time,
                    season={x[0]: x[1] for x in event["season"]}
                )

    @property
    def sport(self):
        """ Return LookupSport instance for this event
        """
        return LookupSport(self["sport_identifier"])

    @property
    def teams(self):
        """ Return the list of teams

            ... note:: The first element is the **home** team!

        """
        return self["teams"]

    @property
    def eventgroup(self):
        """ Get the event group that corresponds to this event
        """
        sport = LookupSport(self["sport_identifier"])
        return(LookupEventGroup(
            sport["identifier"],
            self["eventgroup_identifier"]))

    def test_operation_equal(self, event):
        """ This method checks if an object or operation on the blockchain
            has the same content as an object in the  lookup
        """
        lookupnames = self.names
        lookupseason = self.season
        chainsnames = [[]]
        chainseason = [[]]
        if "name" in event:
            chainsnames = event["name"]
            event_group_id = event["event_group_id"]
            chainseason = event["season"]
        elif "new_name" in event:
            chainsnames = event["new_name"]
            event_group_id = event["new_sport_id"]
            chainseason = event["new_season"]
        else:
            raise ValueError

        parts = event_group_id.split(".")
        assert len(parts) == 3,\
            "{} is a strange sport object id".format(event_group_id)
        if int(parts[0]) == 0:
            event_group_id = ""

        if (all([a in chainsnames for a in lookupnames]) and
                all([b in lookupnames for b in chainsnames]) and
                (lookupseason and  # only test if a season is provided
                    all([b in lookupseason for b in chainseason]) and
                    all([b in chainseason for b in lookupseason])) and
                (not event_group_id or self.parent_id == event_group_id)):
            return True

    def find_id(self):
        """ Try to find an id for the object of the  lookup on the
            blockchain

            ... note:: This only checks if a sport exists with the same name in
                       **ENGLISH**!
        """
        # In case the parent is a proposal, we won't
        # be able to find an id for a child
        parent_id = self.parent.id
        if parent_id[0] == "0" or parent_id[:4] == "1.10":
            return

        events = Events(
            self.parent_id,
            peerplays_instance=self.peerplays)

        # FIXME: Might be we also need to look for season
        en_descrp = next(filter(lambda x: x[0] == "en", self.names))

        for event in events:
            if en_descrp in event["name"]:
                return event["id"]

    def is_synced(self):
        """ Test if data on chain matches lookup
        """
        if "id" in self and self["id"]:
            event = Event(self["id"])
            if self.test_operation_equal(event):
                return True
        return False

    def propose_new(self):
        """ Propose operation to create this object
        """
        return self.peerplays.event_create(
            self.names,
            self.season,
            self["start_time"],
            event_group_id=self.parent_id,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    def propose_update(self):
        """ Propose to update this object to match  lookup
        """
        return self.peerplays.event_update(
            self["id"],
            self.names,
            self.season,
            self["start_time"],
            event_group_id=self.parent_id,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    @property
    def participants(self):
        """ Return content of participants in this event
        """
        from .participant import LookupParticipants
        name = self.eventgroup["participants"]
        return LookupParticipants(
            self["sport_identifier"], name)

    def lookup_bettingmarketgroups(self):
        """ Return content of betting market groups
        """
        names = self.eventgroup["bettingmarketgroups"]
        for name in names:
            yield self.eventgroup.sport["bettingmarketgroups"][name]

    @property
    def names(self):
        """ Properly format names for internal use
        """
        return [
            [
                k,
                v
            ] for k, v in self.names_json.items()
        ]

    @property
    def names_json(self):
        """ This property derives the names for each language provided in the
            eventscheme and fills in the variables.

            :rtype dict
        """
        teams = self["teams"]
        scheme = self.eventscheme.get("name", {})
        return subsitution(teams, scheme)

    @property
    def season(self):
        """ Properly format season for internal use
        """
        return [
            [
                k,
                v
            ] for k, v in self["season"].items()
        ]

    @property
    def eventscheme(self):
        """ Obtain Event scheme from event group
        """
        return self.eventgroup["eventscheme"]

    @property
    def bettingmarketgroups(self):
        """ Return instances of LookupBettingMarketGroup for this event
        """
        for bmg in self.lookup_bettingmarketgroups():
            yield LookupBettingMarketGroup(bmg, event=self)



#!/usr/bin/env python

from setuptools import setup
import sys

__VERSION__ = '0.0.1'

assert sys.version_info[0] == 3, "We require Python > 3"

setup(
    name='bookied-sync',
    version=__VERSION__,
    description=(
        'Synchronized objects on the blockchain using proposals where'
        'necessary'
    ),
    long_description=open('README.md').read(),
    download_url='https://github.com/pbsa/bookied/tarball/' + __VERSION__,
    author='Fabian Schuh',
    author_email='Fabian.Schuh@BlockchainProjectsBV.com',
    maintainer='Fabian Schuh',
    maintainer_email='Fabian.Schuh@BlockchainProjectsBV.com',
    url='http://pbsa.info',
    keywords=['peerplays', 'bookie'],
    packages=[
        "bookied_sync"
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
    ],
    entry_points={
        'console_scripts': [
            'bookied = bookie.cli:main'
        ],
    },
    install_requires=[
        "peerplays",
        "pyyaml",
        "colorlog",
        "dateutils",
        "bookiesports==0.0.2",
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    include_package_data=True,
)



from peerplays.sport import Sports, Sport
from .lookup import Lookup
from .eventgroup import LookupEventGroup
from .rule import LookupRules
from .bettingmarketgroup import LookupBettingMarketGroup
from .participant import LookupParticipants
from .exceptions import ObjectNotFoundInLookup


class LookupSport(Lookup, dict):
    """ Lookup Class for a Sport

        :param str sport: Identifier for the Sport

    """

    operation_update = "sport_update"
    operation_create = "sport_create"

    def __init__(self, sport):
        self.identifier = sport
        super(LookupSport, self).__init__()

        if sport.lower() in [x.lower() for x in self.data["sports"]]:
            # Easy, the sports name is the key
            dict.__init__(self, self.data["sports"][sport])
        else:
            found = False
            # Load from identifier
            for name, s in self.data["sports"].items():
                if (
                    # Name
                    name.lower() == sport.lower() or
                    # Identifier
                    s.get("identifier", "").lower() == sport.lower() or
                    # List of languages
                    sport.lower() in [
                        x.lower()for x in s.get("name", {}).values()] or
                    # List of aliases
                    sport.lower() in [
                        x.lower() for x in s.get("aliases", [])]
                ):
                    found = True
                    dict.__init__(self, s)

            if not found:
                raise ObjectNotFoundInLookup("Not Found: {}".format(
                    sport))

    @property
    def eventgroups(self):
        """ Return instances of LookupEventGroup for all event groups in this
            sport
        """
        for e in self["eventgroups"]:
            yield LookupEventGroup(
                self.identifier, e)

    @property
    def rules(self):
        """ Return instances of LookupRules for all rules in this sport
        """
        for e in self["rules"]:
            yield LookupRules(
                self.identifier, e)

    @property
    def participants(self):
        """ Return instances of LookupParticipants for each participant in this
            sport
        """
        for e in self["participants"]:
            yield LookupParticipants(
                self.identifier, e)

    @property
    def bettingmarketgroups(self):
        """ Return isntances of LookupBettingMarketGroup for each betting market
            group of this sport
        """
        for e in self["bettingmarketgroups"]:
            yield LookupBettingMarketGroup(
                self.identifier, e)

    def test_operation_equal(self, sport):
        """ This method checks if an object or operation on the blockchain
            has the same content as an object in the  lookup
        """
        lookupnames = self.names
        chainsnames = [[]]
        if "name" in sport:
            chainsnames = sport["name"]
        elif "new_name" in sport:
            chainsnames = sport["new_name"]
        else:
            raise ValueError
        if (all([a in chainsnames for a in lookupnames]) and
                all([b in lookupnames for b in chainsnames])):
            return True

    def find_id(self):
        """ Try to find an id for the object of the  lookup on the
            blockchain

            ... note:: This only checks if a sport exists with the same name in
                       **ENGLISH**!
        """
        sports = Sports(peerplays_instance=self.peerplays)
        en_descrp = next(filter(lambda x: x[0] == "en", self.names))
        for sport in sports:
            if en_descrp in sport["name"]:
                return sport["id"]

    def is_synced(self):
        """ Test if data on chain matches lookup
        """
        if "id" in self:
            sport = Sport(self["id"])
            if self.test_operation_equal(sport):
                return True
        return False

    def propose_new(self):
        """ Propose operation to create this object
        """
        return self.peerplays.sport_create(
            self.names,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer)

    def propose_update(self):
        """ Propose to update this object to match  lookup
        """
        return self.peerplays.sport_update(
            self["id"],
            names=self.names,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer)

    @property
    def names(self):
        """ Properly format names for internal use
        """
        return [
            [
                k,
                v
            ] for k, v in self["name"].items()
        ]



import re
import sys
from peerplays.eventgroup import EventGroups, EventGroup
from .lookup import Lookup, LookupDatabaseConfig
from .exceptions import ObjectNotFoundInLookup
from . import log


class LookupEventGroup(Lookup, dict):
    """ Lookup Class for Event Group

        :param str sport: Identifier of the Sport
        :param str eventgroup: Identifier of the Eventgroup

    """

    operation_update = "event_group_update"
    operation_create = "event_group_create"

    def __init__(self, sport, eventgroup):
        from .sport import LookupSport
        self.identifier = eventgroup

        if isinstance(sport, LookupSport):
            sport = sport.get("identifier")

        self.sport_name = sport
        self.sport = LookupSport(sport)
        self.parent = self.sport
        self.eventgroup = eventgroup
        super(LookupEventGroup, self).__init__()

        if eventgroup.lower() in self.data["sports"][sport]["eventgroups"]:
            dict.__init__(
                self,
                self.data["sports"][sport]["eventgroups"][eventgroup]
            )
        else:
            found = False
            for name, evg in self.data["sports"][sport]["eventgroups"].items():
                if (
                    # Name
                    name.lower() == eventgroup.lower() or
                    # Identifier
                    evg.get("identifier", "").lower() == eventgroup.lower() or
                    # List of languages
                    eventgroup.lower() in [
                        x.lower()for x in evg.get("name", {}).values()] or
                    # List of aliases
                    eventgroup.lower() in [
                        x.lower() for x in evg.get("aliases", [])]
                ):
                    found = True
                    dict.__init__(self, evg)

            if not found:
                raise ObjectNotFoundInLookup(
                    "Eventgroup {} not avaialble in sport {}".format(
                        eventgroup, sport))

    def test_operation_equal(self, eventgroup):
        """ This method checks if an object or operation on the blockchain
            has the same content as an object in the  lookup
        """
        lookupnames = self.names
        chainsnames = [[]]
        if "name" in eventgroup:
            chainsnames = eventgroup["name"]
            sport_id = eventgroup["sport_id"]
        elif "new_name" in eventgroup:
            chainsnames = eventgroup["new_name"]
            sport_id = eventgroup["new_sport_id"]
        else:
            raise ValueError

        parts = sport_id.split(".")
        assert len(parts) == 3,\
            "{} is a strange sport object id".format(sport_id)
        if int(parts[0]) == 0:
            sport_id = ""

        if (all([a in chainsnames for a in lookupnames]) and
                all([b in lookupnames for b in chainsnames]) and
                (not sport_id or self.parent.id == sport_id)):
            return True

    def find_id(self):
        """ Try to find an id for the object of the  lookup on the
            blockchain

            ... note:: This only checks if a sport exists with the same name in
                       **ENGLISH**!
        """
        # In case the parent is a proposal, we won't
        # be able to find an id for a child
        parent_id = self.parent.id
        if parent_id[0] == "0" or parent_id[:4] == "1.10":
            return

        egs = EventGroups(
            self.parent.id,
            peerplays_instance=self.peerplays)
        en_descrp = next(filter(lambda x: x[0] == "en", self.names))

        for eg in egs:
            if (
                en_descrp in eg["name"] and
                self.parent.id == eg["sport_id"]
            ):
                return eg["id"]

    def is_synced(self):
        """ Test if data on chain matches lookup
        """
        if "id" in self and self["id"]:
            eventgroup = EventGroup(self["id"])
            if self.test_operation_equal(eventgroup):
                return True
        return False

    def propose_new(self):
        """ Propose operation to create this object
        """
        return self.peerplays.event_group_create(
            self.names,
            sport_id=self.parent_id,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    def propose_update(self):
        """ Propose to update this object to match  lookup
        """
        return self.peerplays.event_group_update(
            self["id"],
            names=self.names,
            sport_id=self.parent_id,
            account=self.proposing_account,
            append_to=Lookup.proposal_buffer
        )

    @property
    def events(self):
        """ Return all events that our local database believes correspond to
            this event group
        """
        try:
            from bookied_scrapers.sinks.mysql_data_sink import MysqlDataSink
            from bookied_scrapers.sinks.db_data_sink_model import RawGameInfo, RawEvent, RawMarketsInfo, ResolvedGameEvent
            from playhouse.shortcuts import model_to_dict
        except ImportError:
            raise ImportError("Please install bookied-scrapers")
        from .event import LookupEvent

        MysqlDataSink(
            LookupDatabaseConfig.name,
            LookupDatabaseConfig.user,
            LookupDatabaseConfig.password)
        events = (
            ResolvedGameEvent.select(
                ResolvedGameEvent
            ).where(
                (ResolvedGameEvent.game == self.sport["identifier"]) &
                (ResolvedGameEvent.league == self.eventgroup)
            )
        )
        for e in events:
            event = model_to_dict(e)
            # Replace id since mysql uses different ids
            event["id"] = None
            teams = re.split(r"[:@]", event["teams"])
            yield LookupEvent(
                name={"en": event["teams"]},
                teams=[t.strip() for t in teams],
                eventgroup_identifier=event["league"],
                sport_identifier=event["game"],
                season={},
                start_time=event["start_time"],
                extra_data=event,
            )

    @property
    def names(self):
        """ Properly format names for internal use
        """
        return [
            [
                k,
                v
            ] for k, v in self["name"].items()
        ]
